/*

 XPressNet TCO (TABLERO DE CONTROL OPTICO)
 Version 1.0
 
 Copyright (C) 2010 John Alexander Stewart
 Copyright (C) 2015 Norberto Redondo Melchor
 
 Para Arduinos basados en el ATMega328 como el Uno o el Pro Mini 
 */

#define MY_ADDRESS 5                // DIRECCION EN EL BUS XPressNet PARA RS-485 transmission. 
// AQUI ASOCIO NUM DE ACCESORIO CON SU PULSADOR Y SU LED INDICADOR:
int correspondencias[][4]={      // { xNum accesorio en Xbus, su pinPulsador, su pinLed, su estActual}
  {102, 3, 4, LOW},
  {103, 5, 6, LOW},
  {23, 7, 8, LOW},
  {21, 9, 10, LOW},
  {22, 11, 12, LOW},
};
int cantParesAccLed;      // Cantidad de pares accesorio & led asociado

#define MAX485_CONTROL 2            //Arduino Pin for send or receive control
#define LISTEN_MODE digitalWrite (MAX485_CONTROL, LOW)
#define SEND_MODE digitalWrite (MAX485_CONTROL, HIGH)

#define LED_ON digitalWrite (13, HIGH)              // SEÑALIZA CENTRAL OK
#define LED_OFF digitalWrite (13, LOW)

#define LEDXbus 14                                  // SEÑALIZA TRANSMISION XPRESSNET.  El pin 14 es el A0
#define LEDXbus_ON digitalWrite (LEDXbus, HIGH)  
#define LEDXbus_OFF digitalWrite (LEDXbus, LOW)


#define WAIT_FOR_XMIT_COMPLETE {while (!(UCSR0A & (1<<TXC0))); UCSR0A = (1<<TXC0); UCSR0A = 0;}
#define GENERAL_BROADCAST 0x160
unsigned int myDirectedOps;
unsigned int myCallByteInquiry;
unsigned int myRequestAck;

#define MAXSAVED 20
int dataSaved;                                   // for saving incoming data, we use these 2 variables.
int savedData[MAXSAVED];

boolean emergencyStop = false;                   // certain global XPressnet status indicators
boolean trackPowerOff = false;
boolean inServiceMode = false;

int displayLEDState = LOW;                       // for the status LED;   
long flashInterval = 300;
long currentFlashCounter = 0;

#define GET_COMMAND_STATION_STATUS_STEP 0        // power up initalization variables
#define GET_COMMAND_STATION_SW_VERSION_STEP 1
#define WAIT_FOR_COMMAND_STATION_POWERUP 3
#define PAST_INITIALIZATION 200
int initializationStep = GET_COMMAND_STATION_STATUS_STEP;
int initializeError = 0;  // flashes LED pattern on error

unsigned char commandStatusSequence[] = {  0x21, 0x24, 0x05};
unsigned char commandSWVersionSequence[] = {  0x21,0x21,0x00};
unsigned char requestAckAck[] = {  0x20, 0x20};
unsigned char accDecInfoRequest[] = {  0x42, 0x00, 0x00, 0x00};            // Preguntar estado accesorio
unsigned char accDecOpsCommand[] = {  0x52, 0x00, 0x00, 0x00};             // Ordenar cambio estado accesorio
unsigned int numDesvioPar = 0;             // La CS informa siempre sobre dos desviosa la vez,
unsigned int numDesvioImpar = 0;           // el par (0 o 2) y el impar (1 o 3) de cada grupo de 4 (la misma xAddress)
unsigned int statusDesvioPar = 0;
unsigned int statusDesvioImpar = 0;
unsigned int refrescando = 0;
unsigned int accNum;
unsigned int resto;

#define TASA_DE_REFRESCO 3000           // Variables auxiliares
#define PAUSA_ENTRE_COMANDOS 50         // En el bus se piden las cosas poco a poco!
#define CONT_REBOTE 250                 // Se requiere una buena pulsacion!
int contPulsador = 0;
int contRefrescoTCO = 0; 
int contPausa = 0;
boolean switchControlsChanged = false;  // Para ejecutar nueva pulsacion
boolean tocaDesactivar = false;         // Para desactivar cada comando activado
boolean hablarAlgo = false;             // Para cuando quiera hablarle a la central
boolean refreshNow = false;             // Toca refrescar leds en panel


void xpressnetSpeed () {                //Set up serial port on 62500 Baud
  UBRR0H = 0;
  UBRR0L = 0x0F;
  UCSR0A = 0;
  UCSR0B = (1<<RXEN0) | (1<<TXEN0) | (0<<RXCIE0) | (1<<UCSZ02);
  UCSR0C = (1<<UCSZ01) | (1<<UCSZ00);
}


unsigned int callByteParity (unsigned int me) {  // calculate the parity bit in the call byte for this guy
  int parity = (1==0);
  unsigned int vv;
  me = me & 0x7f;
  vv = me;
  while (vv) {
    parity = !parity;
    vv = vv & (vv-1);
  }
  if (parity) me = me | 0x80;
  return me;
}


void setup() {
  cantParesAccLed = 5;                // En general habria que poner   = sizeof(correspondencias)/sizeof(int);

  pinMode(LEDXbus, OUTPUT);           // Indicador transmision por Xbus en sendCommand()

  pinMode(MAX485_CONTROL, OUTPUT);    // the MAX485 chip for Xpressnet - set this up.
  LISTEN_MODE;
  LED_OFF;       // turn the LED off until we are running

  int i;
  for (i=0; i < cantParesAccLed; i++) {  //  CONFIGURAMOS PULSADORES Y ENCENDEMOS LOS LEDS UTILIZADOS
    pinMode(correspondencias[i][1], INPUT_PULLUP);
    pinMode(correspondencias[i][2], OUTPUT);
    digitalWrite(correspondencias[i][2], HIGH);
    delay(10);
  }
  delay(500);
  for (i=0; i < cantParesAccLed; i++) {  // APAGAMOS LOS LEDS UTILIZADOS
    digitalWrite(correspondencias[i][2], LOW);
  }
  dataSaved = 0;
  myRequestAck = callByteParity (MY_ADDRESS | 0x00) | 0x100;   // set these up to match call bytes for this device
  myCallByteInquiry = callByteParity (MY_ADDRESS | 0x40) | 0x100;
  myDirectedOps = callByteParity (MY_ADDRESS | 0x60) | 0x100;
  xpressnetSpeed();                                           // AJUSTAR EL PUERTO SERIE A 1B+9B+1STOP
}


int USART_Receive( void ) {
  unsigned char status, resh, resl;
  status = UCSR0A;
  while ( !(status & (1<<RXC0)) ) {
    status = UCSR0A;
  }
  resh = UCSR0B;
  resl = UDR0;
  if ( status & ((1<<FE0)|(1<<DOR0)|(1<<UPE0))) { 
    return -1;
  }
  resh = (resh >> 1) & 0x01;
  return ((resh << 8) | resl);
}


void USART_Transmit (unsigned char data8) {
  while (!(UCSR0A & (1<<UDRE0))) {
  }
  UDR0 = data8;
}


void sendCommand(unsigned char *dataString, int byteCount) {
  LEDXbus_ON;
  unsigned int i;
  SEND_MODE;
  delayMicroseconds(20);
  for (i=0; i< byteCount; i++) {
    USART_Transmit (*dataString);
    dataString ++;
  }
  WAIT_FOR_XMIT_COMPLETE;
  LISTEN_MODE;
  LEDXbus_OFF;
}  


void calculateXOR (unsigned char *dataString, int byteCount) { // make up the XOR byte, and put it in the LAST position.
  unsigned int i;
  unsigned char xo;
  xo = 0;
  for (i=0; i<(byteCount-1); i++) {
    xo = xo ^ (*dataString);
    dataString ++;
  }
  *dataString = xo;
}


void loopErrorToLED(int pattern) {   // on a catastrophic error, loop until reset, flashing...
  int i;
  for (;;) {
    for (i=0; i<pattern; i++) {
      LED_ON;
      delay (100);
      LED_OFF;
      delay (100);
    }
    delay (1000);
  }
}


void handle_possible_track_power_indications() {    
  if (trackPowerOff) {
    currentFlashCounter++;
    if (currentFlashCounter >= flashInterval) {
      currentFlashCounter = 0;
      displayLEDState = !displayLEDState;
      if (displayLEDState) {
        LED_ON;
      } 
      else {
        LED_OFF;
      }
    }
  } 
  else {
    LED_ON;  // we are in running mode...
  }     
}


void handleSavedData() {    // AQUI DISCRIMINA SI SE HA RECIBIDO MENSAJE GENERAL O PARA NOSOTROS
  if (*savedData == GENERAL_BROADCAST) {
    handleGeneralBroadcast();
  } 
  else if (*savedData == myDirectedOps) {
    handleDirectedOps();
  }
}


void handleGeneralBroadcast () {   // Data destined to ALL XpressNet devices; track power off, etc...
  int i;
  if (dataSaved == 4) {
    if (savedData[1] == 0x61) {
      if ((savedData[2] == 0x01) && (savedData[3] == 0x60)) {
        emergencyStop = false;
        trackPowerOff = false;
        inServiceMode = false;
        return;
      } 
      else if ((savedData[2] == 0x00) && (savedData[3] == 0x61)) {
        trackPowerOff = true;
        return;
      } 
      else if ((savedData[2] == 0x02) && (savedData[3] == 0x63)) {
        inServiceMode = true;
        return;
      }
    } 
    else if (savedData[1] == 0x81) {
      if ((savedData[2] == 0x00) && (savedData[3] == 0x81)) {
        emergencyStop = true;
        return;
      }
    }
  } 
  else if (dataSaved == 8) {
    if (savedData[1] == 0x05) { //  DCC FAST CLOCK set request
      return;
    }
  }
}


void handleDirectedOps () {       // AQUI INTERPRETAMOS LOS MENSAJES QUE NOS LLEGAN POR EL XBUS
  int i;
  unsigned char dataByte1;
  unsigned char dataByte2;
#define DATA_INDICATOR savedData[1]
  if (dataSaved >= 2) {
    dataByte1 = savedData[2];
    dataByte2 = savedData[3];
    if (DATA_INDICATOR == 0x62) {
      if (dataByte1 == 0x21) {                  // Version 2.3
        loopErrorToLED(5);
      } 
      else if (dataByte1 == 0x22) {             // CS status response, see section 2.1.7
        emergencyStop = false;
        trackPowerOff = false;
        inServiceMode = false;
        if (dataByte2 != 0) {
          if ((dataByte2 & 0x01) == 0x01) {     // is track power turned off?
            trackPowerOff = true;
          }
          if ((dataByte2 & 0x02) == 0x02) {     // is it in emergency stop?
            emergencyStop = true;
          }
          if ((dataByte2 & 0x08) == 0x08) {     // in service mode?
            inServiceMode = true;
          }             
          if ((dataByte2 & 0x40) == 0x40) {    // in powerup mode - wait until complete
            initializationStep = WAIT_FOR_COMMAND_STATION_POWERUP;
          }
        }
        if (initializationStep == GET_COMMAND_STATION_STATUS_STEP) {
          initializationStep = GET_COMMAND_STATION_SW_VERSION_STEP;
        }
      }
    } 
    else if (DATA_INDICATOR  == 0x63) {      // CS software version
      if (dataByte1 == 0x21) {
        if (savedData[3] < 0x30) loopErrorToLED(5);
        if (initializationStep == GET_COMMAND_STATION_SW_VERSION_STEP) {
          initializationStep = PAST_INITIALIZATION;
        }
      }
    } 
    else if (DATA_INDICATOR == 0x42) {      // Ha llegado un 0x42,dataByte1,dataByte2: es estado de accesorio
      estadoAccesorios(dataByte1, dataByte2);
      for (i=0; i < cantParesAccLed; i++) {    // Recorro todas las asociaciones para encontrar la necesaria; i podria correr de dos en dos
        if (correspondencias[i][0] == numDesvioPar) {
          digitalWrite(correspondencias[i][2], statusDesvioPar & 0x01);      // Enciendo el led del desvioPar
          correspondencias[i][3] = statusDesvioPar & 0x01;
        }
        if (correspondencias[i][0] == numDesvioImpar) {
          digitalWrite(correspondencias[i][2], statusDesvioImpar & 0x01);      // Enciendo el led del desvioPar
          correspondencias[i][3] = statusDesvioImpar & 0x01;
        }        
      }
    }
  }
}


void estadoAccesorios(int Byte1, int Byte2) {  // Para extraer numDesvio, statusDesvio descifrando ITTN ZZZZ
  /* Byte1 es Address  Byte2 es ITTN ZZZZ
   When Address = 0x00, (turnout group 0) the information is about turnouts 0,  1,  2   3 
   Info in Byte2 = I T T N - Z3 Z2 Z1 Z0 :
   I:       0 = accesorio Ok   1 = accesorio cambiando estado
   TT:     00 = accesorio sin feedback
   TT:     01 = accesorio con feedback
   TT:     10 = detector de consumo en retromodulo
   TT:     11 = no usado
   N:       0 = lower nibble  (Z0,Z1 desvio par 0 y Z2,Z3 desvio impar 1)
   N:       1 = uper nible    (Z0,Z1 desvio par 2 y Z2,Z3 desvio impar 3)
   ZZ:     00 = no controlado aun
   ZZ:     01 = arriba en el Multimaus (1)
   ZZ:     10 = abajo en el Multimaus  (2)
   ZZ:     11 = error
   */
  int N = (Byte2 & 0x10) >> 3;              // 0x10 = B 0001 0000
  numDesvioPar = (Byte1 << 2) + N;         // es el xNum; el mNum es uno mas
  numDesvioImpar = numDesvioPar + 1;
  statusDesvioPar = Byte2 & 0x03;           // desvios 0 y 2; status puede ser 00, 01, 10, 11
  statusDesvioImpar = (Byte2 >> 2) & 0x03;  // 0x03 = B 0000 0011
}


void read_switches() {	
  int i;
  for (i=0; i < cantParesAccLed; i++) {    // Recorro todas las asociaciones buscando pulsaciones
    if (digitalRead(correspondencias[i][1]) == LOW  && tocaDesactivar == false) {    // El pulsador esta activado al ponerse a GND
      contPulsador ++;
      if (contPulsador == CONT_REBOTE) {
        accNum = correspondencias[i][0];
        accDecOpsCommand[1] = accNum >> 2;      // xAddress = Address / 4
        resto = accNum % 4;
        accDecOpsCommand[2] = 0x80 + (resto << 1) + correspondencias[i][3];     
        // 0x80 y 0x81 afectan al desvio x20 del Xbus (el m21 del Multim)
        // 0x84 y 0x85 afectan al desvio x22 del Xbus (el m23 del Multim)
        calculateXOR(accDecOpsCommand, 4);
        correspondencias[i][3] = correspondencias[i][3] ^ 0x01;  // Memorizo estado opuesto
        switchControlsChanged = true;
        tocaDesactivar = true;
        contPulsador = 0;
        return;
      }
    }
  }  
  if (switchControlsChanged == false && tocaDesactivar == true) {    // Desactivo salida para que el comando surta efecto
    accDecOpsCommand[2] = accDecOpsCommand[2] ^ 0x08;                // 0x08 = B 0000 1000
    calculateXOR(accDecOpsCommand, 4);
    switchControlsChanged = true;
    tocaDesactivar = false;
    pideInfo(accNum);
  }
}

void pideInfo(int xNum) {
    int rest = xNum % 4;
    accDecInfoRequest[1] = xNum >> 2;
    accDecInfoRequest[2] = 0x80 + (rest >> 1);
    calculateXOR(accDecInfoRequest, 4);
    hablarAlgo = true;
}


void periodic_Refresh_TCO() {
  contRefrescoTCO++;
  if (contRefrescoTCO > TASA_DE_REFRESCO) {
    contPausa++;
    if (contPausa > PAUSA_ENTRE_COMANDOS) {
      if (refrescando < cantParesAccLed) {
        pideInfo(correspondencias[refrescando][0]);
        refrescando++;
        contPausa = 0;
      }
      else {
        refrescando = 0;
        contRefrescoTCO = 0;
     }
    }
  }
}


void handleCallByteInquiry () {                       // Nos ha llegado el turno de comunicar. ¿Mandamos algo?
  if (initializationStep != PAST_INITIALIZATION) {
    if ((initializationStep == GET_COMMAND_STATION_STATUS_STEP) ||
      (initializationStep == WAIT_FOR_COMMAND_STATION_POWERUP)) {
      sendCommand (commandStatusSequence, 3);    
      if (initializationStep == WAIT_FOR_COMMAND_STATION_POWERUP) {
        initializationStep = GET_COMMAND_STATION_STATUS_STEP;
      }
    } 
    else if (initializationStep == GET_COMMAND_STATION_SW_VERSION_STEP) {
      sendCommand (commandSWVersionSequence,3);
    } 
  } 
  else {                                   // WE ARE RUNNING, AND OUR WINDOW IS OPEN! YEEEEHAH!
    if (switchControlsChanged) {
      sendCommand (accDecOpsCommand, 4);   // PONGO 4 PORQUE EL COMANDO ES DE 4 BYTES
      switchControlsChanged = false;
      return;
    } 
    else if (hablarAlgo) {
      sendCommand (accDecInfoRequest, 4);   // PONGO 4 PORQUE EL COMANDO ES DE 4 BYTES
      hablarAlgo = false;
    }
  }
}


void loop() {
  int rxdata = USART_Receive();
  if (rxdata != -1) {
    if (rxdata >= 0x100) {
      if (rxdata == myCallByteInquiry) {
        handleCallByteInquiry();        
      } 
      else if (rxdata == myRequestAck) {
        sendCommand (requestAckAck,2);
      }            
      if (dataSaved > 1) {
        handleSavedData();
        dataSaved = 0;
      }
      if ((rxdata == GENERAL_BROADCAST) ||
        (rxdata == myDirectedOps)) {           
        savedData[0] = rxdata;
        dataSaved=1;       
      } 
    } 
    else {
      if (dataSaved !=0) {
        if (dataSaved < MAXSAVED) {
          savedData[dataSaved] = rxdata;
          dataSaved++;
        }
      }
    }    
  }
  if (initializationStep == PAST_INITIALIZATION) {   
    handle_possible_track_power_indications();
    read_switches();
    periodic_Refresh_TCO();
  } 
}




